@@ -34,45 +34,51 @@
 /*
  * assemble.c   code generation for the Netwide Assembler
  *
- * the actual codes (C syntax, i.e. octal):
- * \0            - terminates the code. (Unless it's a literal of course.)
- * \1..\4        - that many literal bytes follow in the code stream
- * \5            - add 4 to the primary operand number (b, low octdigit)
- * \6            - add 4 to the secondary operand number (a, middle octdigit)
- * \7            - add 4 to both the primary and the secondary operand number
- * \10..\13      - a literal byte follows in the code stream, to be added
- *                 to the register value of operand 0..3
- * \14..\17      - the position of index register operand in MIB (BND insns)
- * \20..\23      - a byte immediate operand, from operand 0..3
- * \24..\27      - a zero-extended byte immediate operand, from operand 0..3
- * \30..\33      - a word immediate operand, from operand 0..3
- * \34..\37      - select between \3[0-3] and \4[0-3] depending on 16/32 bit
- *                 assembly mode or the operand-size override on the operand
- * \40..\43      - a long immediate operand, from operand 0..3
- * \44..\47      - select between \3[0-3], \4[0-3] and \5[4-7]
- *                 depending on the address size of the instruction.
- * \50..\53      - a byte relative operand, from operand 0..3
- * \54..\57      - a qword immediate operand, from operand 0..3
- * \60..\63      - a word relative operand, from operand 0..3
- * \64..\67      - select between \6[0-3] and \7[0-3] depending on 16/32 bit
- *                 assembly mode or the operand-size override on the operand
- * \70..\73      - a long relative operand, from operand 0..3
- * \74..\77      - a word constant, from the _segment_ part of operand 0..3
- * \1ab          - a ModRM, calculated on EA in operand a, with the spare
- *                 field the register value of operand b.
- * \172\ab       - the register number from operand a in bits 7..4, with
- *                 the 4-bit immediate from operand b in bits 3..0.
- * \173\xab      - the register number from operand a in bits 7..4, with
- *                 the value b in bits 3..0.
- * \174..\177    - the register number from operand 0..3 in bits 7..4, and
- *                 an arbitrary value in bits 3..0 (assembled as zero.)
- * \2ab          - a ModRM, calculated on EA in operand a, with the spare
- *                 field equal to digit b.
+ * Bytecode specification
+ * ----------------------
+ *
+ *
+ * Codes            Mnemonic        Explanation
+ *
+ * \0                                       terminates the code. (Unless it's a literal of course.)
+ * \1..\4                                   that many literal bytes follow in the code stream
+ * \5                                       add 4 to the primary operand number (b, low octdigit)
+ * \6                                       add 4 to the secondary operand number (a, middle octdigit)
+ * \7                                       add 4 to both the primary and the secondary operand number
+ * \10..\13                                 a literal byte follows in the code stream, to be added
+ *                                          to the register value of operand 0..3
+ * \14..\17                                 the position of index register operand in MIB (BND insns)
+ * \20..\23         ib                      a byte immediate operand, from operand 0..3
+ * \24..\27         ib,u                    a zero-extended byte immediate operand, from operand 0..3
+ * \30..\33         iw                      a word immediate operand, from operand 0..3
+ * \34..\37         iwd                     select between \3[0-3] and \4[0-3] depending on 16/32 bit
+ *                                          assembly mode or the operand-size override on the operand
+ * \40..\43         id                      a long immediate operand, from operand 0..3
+ * \44..\47         iwdq                    select between \3[0-3], \4[0-3] and \5[4-7]
+ *                                          depending on the address size of the instruction.
+ * \50..\53         rel8                    a byte relative operand, from operand 0..3
+ * \54..\57         iq                      a qword immediate operand, from operand 0..3
+ * \60..\63         rel16                   a word relative operand, from operand 0..3
+ * \64..\67         rel                     select between \6[0-3] and \7[0-3] depending on 16/32 bit
+ *                                          assembly mode or the operand-size override on the operand
+ * \70..\73         rel32                   a long relative operand, from operand 0..3
+ * \74..\77         seg                     a word constant, from the _segment_ part of operand 0..3
+ * \1ab                                     a ModRM, calculated on EA in operand a, with the spare
+ *                                          field the register value of operand b.
+ * \172\ab                                  the register number from operand a in bits 7..4, with
+ *                                          the 4-bit immediate from operand b in bits 3..0.
+ * \173\xab                                 the register number from operand a in bits 7..4, with
+ *                                          the value b in bits 3..0.
+ * \174..\177                               the register number from operand 0..3 in bits 7..4, and
+ *                                          an arbitrary value in bits 3..0 (assembled as zero.)
+ * \2ab                                     a ModRM, calculated on EA in operand a, with the spare
+ *                                          field equal to digit b.
+ *
+ * \240..\243                               this instruction uses EVEX rather than REX or VEX/XOP, with the
+ *                                          V field taken from operand 0..3.
+ * \250                                     this instruction uses EVEX rather than REX or VEX/XOP, with the
+ *                                          V field set to 1111b.
  *
- * \240..\243    - this instruction uses EVEX rather than REX or VEX/XOP, with the
- *                 V field taken from operand 0..3.
- * \250          - this instruction uses EVEX rather than REX or VEX/XOP, with the
- *                 V field set to 1111b.
  * EVEX prefixes are followed by the sequence:
  * \cm\wlp\tup    where cm is:
  *                  cc 000 0mm
@@ -97,11 +103,11 @@
  *                tup is tuple type for Disp8*N from %tuple_codes in insns.pl
  *                    (compressed displacement encoding)
  *
- * \254..\257    - a signed 32-bit operand to be extended to 64 bits.
- * \260..\263    - this instruction uses VEX/XOP rather than REX, with the
- *                 V field taken from operand 0..3.
- * \270          - this instruction uses VEX/XOP rather than REX, with the
- *                 V field set to 1111b.
+ * \254..\257       id,s                        a signed 32-bit operand to be extended to 64 bits.
+ * \260..\263                                   this instruction uses VEX/XOP rather than REX, with the
+ *                                              V field taken from operand 0..3.
+ * \270                                         this instruction uses VEX/XOP rather than REX, with the
+ *                                              V field set to 1111b.
  *
  * VEX/XOP prefixes are followed by the sequence:
  * \tmm\wlp        where mm is the M field; and wlp is:
@@ -117,62 +123,63 @@
  *
  * t = 0 for VEX (C4/C5), t = 1 for XOP (8F).
  *
- * \271		 - instruction takes XRELEASE (F3) with or without lock
- * \272		 - instruction takes XACQUIRE/XRELEASE with or without lock
- * \273          - instruction takes XACQUIRE/XRELEASE with lock only
- * \274..\277    - a byte immediate operand, from operand 0..3, sign-extended
- *                 to the operand size (if o16/o32/o64 present) or the bit size
- * \310          - indicates fixed 16-bit address size, i.e. optional 0x67.
- * \311          - indicates fixed 32-bit address size, i.e. optional 0x67.
- * \312          - (disassembler only) invalid with non-default address size.
- * \313          - indicates fixed 64-bit address size, 0x67 invalid.
- * \314          - (disassembler only) invalid with REX.B
- * \315          - (disassembler only) invalid with REX.X
- * \316          - (disassembler only) invalid with REX.R
- * \317          - (disassembler only) invalid with REX.W
- * \320          - indicates fixed 16-bit operand size, i.e. optional 0x66.
- * \321          - indicates fixed 32-bit operand size, i.e. optional 0x66.
- * \322          - indicates that this instruction is only valid when the
- *                 operand size is the default (instruction to disassembler,
- *                 generates no code in the assembler)
- * \323          - indicates fixed 64-bit operand size, REX on extensions only.
- * \324          - indicates 64-bit operand size requiring REX prefix.
- * \325          - instruction which always uses spl/bpl/sil/dil
- * \326          - instruction not valid with 0xF3 REP prefix.  Hint for
-                   disassembler only; for SSE instructions.
- * \330          - a literal byte follows in the code stream, to be added
- *                 to the condition code value of the instruction.
- * \331          - instruction not valid with REP prefix.  Hint for
- *                 disassembler only; for SSE instructions.
- * \332          - REP prefix (0xF2 byte) used as opcode extension.
- * \333          - REP prefix (0xF3 byte) used as opcode extension.
- * \334          - LOCK prefix used as REX.R (used in non-64-bit mode)
- * \335          - disassemble a rep (0xF3 byte) prefix as repe not rep.
- * \336          - force a REP(E) prefix (0xF3) even if not specified.
- * \337          - force a REPNE prefix (0xF2) even if not specified.
- *                 \336-\337 are still listed as prefixes in the disassembler.
- * \340          - reserve <operand 0> bytes of uninitialized storage.
- *                 Operand 0 had better be a segmentless constant.
- * \341          - this instruction needs a WAIT "prefix"
- * \360          - no SSE prefix (== \364\331)
- * \361          - 66 SSE prefix (== \366\331)
- * \364          - operand-size prefix (0x66) not permitted
- * \365          - address-size prefix (0x67) not permitted
- * \366          - operand-size prefix (0x66) used as opcode extension
- * \367          - address-size prefix (0x67) used as opcode extension
- * \370,\371     - match only if operand 0 meets byte jump criteria.
- *                 370 is used for Jcc, 371 is used for JMP.
- * \373          - assemble 0x03 if bits==16, 0x05 if bits==32;
- *                 used for conditional jump over longer jump
- * \374          - this instruction takes an XMM VSIB memory EA
- * \375          - this instruction takes an YMM VSIB memory EA
- * \376          - this instruction takes an ZMM VSIB memory EA
+ * \271             hlexr                       instruction takes XRELEASE (F3) with or without lock
+ * \272             hlenl                       instruction takes XACQUIRE/XRELEASE with or without lock
+ * \273             hle                         instruction takes XACQUIRE/XRELEASE with lock only
+ * \274..\277       ib,s                        a byte immediate operand, from operand 0..3, sign-extended
+ *                                              to the operand size (if o16/o32/o64 present) or the bit size
+ * \310             a16                         indicates fixed 16-bit address size, i.e. optional 0x67.
+ * \311             a32                         indicates fixed 32-bit address size, i.e. optional 0x67.
+ * \312             adf                         (disassembler only) invalid with non-default address size.
+ * \313             a64                         indicates fixed 64-bit address size, 0x67 invalid.
+ * \314             norexb                      (disassembler only) invalid with REX.B
+ * \315             norexx                      (disassembler only) invalid with REX.X
+ * \316             norexr                      (disassembler only) invalid with REX.R
+ * \317             norexw                      (disassembler only) invalid with REX.W
+ * \320             o16                         indicates fixed 16-bit operand size, i.e. optional 0x66.
+ * \321             o32                         indicates fixed 32-bit operand size, i.e. optional 0x66.
+ * \322             odf                         indicates that this instruction is only valid when the
+ *                                              operand size is the default (instruction to disassembler,
+ *                                              generates no code in the assembler)
+ * \323             o64nw                       indicates fixed 64-bit operand size, REX on extensions only.
+ * \324             o64                         indicates 64-bit operand size requiring REX prefix.
+ * \325             nohi                        instruction which always uses spl/bpl/sil/dil
+ * \326             nof3                        instruction not valid with 0xF3 REP prefix.  Hint for
+                                                disassembler only; for SSE instructions.
+ * \330                                         a literal byte follows in the code stream, to be added
+ *                                              to the condition code value of the instruction.
+ * \331             norep                       instruction not valid with REP prefix.  Hint for
+ *                                              disassembler only; for SSE instructions.
+ * \332             f2i                         REP prefix (0xF2 byte) used as opcode extension.
+ * \333             f3i                         REP prefix (0xF3 byte) used as opcode extension.
+ * \334             rex.l                       LOCK prefix used as REX.R (used in non-64-bit mode)
+ * \335             repe                        disassemble a rep (0xF3 byte) prefix as repe not rep.
+ * \336             mustrep                     force a REP(E) prefix (0xF3) even if not specified.
+ * \337             mustrepne                   force a REPNE prefix (0xF2) even if not specified.
+ *                                              \336-\337 are still listed as prefixes in the disassembler.
+ * \340             resb                        reserve <operand 0> bytes of uninitialized storage.
+ *                                              Operand 0 had better be a segmentless constant.
+ * \341             wait                        this instruction needs a WAIT "prefix"
+ * \360             np                          no SSE prefix (== \364\331)
+ * \361                                         66 SSE prefix (== \366\331)
+ * \364             !osp                        operand-size prefix (0x66) not permitted
+ * \365             !asp                        address-size prefix (0x67) not permitted
+ * \366                                         operand-size prefix (0x66) used as opcode extension
+ * \367                                         address-size prefix (0x67) used as opcode extension
+ * \370,\371        jcc8                        match only if operand 0 meets byte jump criteria.
+ *                  jmp8                        370 is used for Jcc, 371 is used for JMP.
+ * \373             jlen                        assemble 0x03 if bits==16, 0x05 if bits==32;
+ *                                              used for conditional jump over longer jump
+ * \374             vsibx|vm32x|vm64x           this instruction takes an XMM VSIB memory EA
+ * \375             vsiby|vm32y|vm64y           this instruction takes an YMM VSIB memory EA
+ * \376             vsibz|vm32z|vm64z           this instruction takes an ZMM VSIB memory EA
  */
 
 #include "compiler.h"
 
 #include <stdio.h>
 #include <string.h>
+#include <stdlib.h>
 #include <inttypes.h>
 
 #include "nasm.h"
@@ -317,7 +324,8 @@ static void out(int64_t offset, int32_t segto, const void *data,
          * convert it into RAWDATA format.
          */
         uint8_t *q = p;
-
+        
+        size = abs((int)size);
         if (size > 8) {
             errfunc(ERR_PANIC, "OUT_ADDRESS with size > 8");
             return;
@@ -345,11 +353,12 @@ static void out(int64_t offset, int32_t segto, const void *data,
     outfmt->output(segto, data, type, size, segment, wrt);
 }
 
-static void out_imm8(int64_t offset, int32_t segment, struct operand *opx)
+static void out_imm8(int64_t offset, int32_t segment,
+                     struct operand *opx, int asize)
 {
     if (opx->segment != NO_SEG) {
         uint64_t data = opx->offset;
-        out(offset, segment, &data, OUT_ADDRESS, 1, opx->segment, opx->wrt);
+        out(offset, segment, &data, OUT_ADDRESS, asize, opx->segment, opx->wrt);
     } else {
         uint8_t byte = opx->offset;
         out(offset, segment, &byte, OUT_RAWDATA, 1, NO_SEG, NO_SEG);
@@ -1323,7 +1332,7 @@ static int64_t calcsize(int32_t segment, int64_t offset, int bits,
             length += 3;
         else
             length += 2;
-    } else if (ins->rex & REX_REAL) {
+    } else if (ins->rex & REX_MASK) {
         if (ins->rex & REX_H) {
             errfunc(ERR_NONFATAL, "cannot use high register in rex instruction");
             return -1;
@@ -1365,10 +1374,10 @@ static int64_t calcsize(int32_t segment, int64_t offset, int bits,
 static inline unsigned int emit_rex(insn *ins, int32_t segment, int64_t offset, int bits)
 {
     if (bits == 64) {
-        if ((ins->rex & REX_REAL) &&
+        if ((ins->rex & REX_MASK) &&
             !(ins->rex & (REX_V | REX_EV)) &&
             !ins->rex_done) {
-            int rex = (ins->rex & REX_REAL) | REX_P;
+            int rex = (ins->rex & REX_MASK) | REX_P;
             out(offset, segment, &rex, OUT_RAWDATA, 1, NO_SEG, NO_SEG);
             ins->rex_done = true;
             return 1;
@@ -1433,7 +1442,7 @@ static void gencode(int32_t segment, int64_t offset, int bits,
                 errfunc(ERR_WARNING | ERR_PASS2 | ERR_WARN_NOV,
                         "byte value exceeds bounds");
             }
-            out_imm8(offset, segment, opx);
+            out_imm8(offset, segment, opx, -1);
             offset += 1;
             break;
 
@@ -1441,7 +1450,7 @@ static void gencode(int32_t segment, int64_t offset, int bits,
             if (opx->offset < 0 || opx->offset > 255)
                 errfunc(ERR_WARNING | ERR_PASS2 | ERR_WARN_NOV,
                         "unsigned byte value exceeds bounds");
-            out_imm8(offset, segment, opx);
+            out_imm8(offset, segment, opx, 1);
             offset += 1;
             break;
 
@@ -1604,7 +1613,7 @@ static void gencode(int32_t segment, int64_t offset, int bits,
                 errfunc(ERR_WARNING | ERR_PASS2 | ERR_WARN_NOV,
                         "signed dword immediate exceeds bounds");
             }
-            out(offset, segment, &data, OUT_ADDRESS, 4,
+            out(offset, segment, &data, OUT_ADDRESS, -4,
                 opx->segment, opx->wrt);
             offset += 4;
             break;
@@ -1881,13 +1890,7 @@ static void gencode(int32_t segment, int64_t offset, int bits,
                 offset += s;
                 s = 0;
 
-                switch (ea_data.bytes) {
-                case 0:
-                    break;
-                case 1:
-                case 2:
-                case 4:
-                case 8:
+                if (ea_data.bytes) {
                     /* use compressed displacement, if available */
                     data = ea_data.disp8 ? ea_data.disp8 : opy->offset;
                     s += ea_data.bytes;
@@ -1900,25 +1903,29 @@ static void gencode(int32_t segment, int64_t offset, int bits,
                                 ea_data.bytes, NO_SEG, NO_SEG);
                         } else {
                             /* overflow check in output/linker? */
-                            out(offset, segment, &data,        OUT_REL4ADR,
+                            out(offset, segment, &data, OUT_REL4ADR,
                                 insn_end - offset, opy->segment, opy->wrt);
                         }
                     } else {
-                        if (overflow_general(data, ins->addr_size >> 3) ||
+                        int asize = ins->addr_size >> 3;
+                        int atype = ea_data.bytes;
+
+                        if (overflow_general(data, asize) ||
                             signed_bits(data, ins->addr_size) !=
-                            signed_bits(data, ea_data.bytes * 8))
+                            signed_bits(data, ea_data.bytes << 3))
                             warn_overflow(ERR_PASS2, ea_data.bytes);
 
+                        if (asize > ea_data.bytes) {
+                            /*
+                             * If the address isn't the full width of
+                             * the address size, treat is as signed...
+                             */
+                            atype = -atype;
+                        }
+
                         out(offset, segment, &data, OUT_ADDRESS,
-                            ea_data.bytes, opy->segment, opy->wrt);
+                            atype, opy->segment, opy->wrt);
                     }
-                    break;
-                default:
-                    /* Impossible! */
-                    errfunc(ERR_PANIC,
-                            "Invalid amount of bytes (%d) for offset?!",
-                            ea_data.bytes);
-                    break;
                 }
                 offset += s;
             }
@@ -2099,6 +2106,31 @@ done:
     return merr;
 }
 
+static uint8_t get_broadcast_num(opflags_t opflags, opflags_t brsize)
+{
+    opflags_t opsize = opflags & SIZE_MASK;
+    uint8_t brcast_num;
+
+    /*
+     * Due to discontinuity between BITS64 and BITS128 (BITS80),
+     * this cannot be a simple arithmetic calculation.
+     */
+    if (brsize > BITS64)
+        errfunc(ERR_FATAL,
+            "size of broadcasting element is greater than 64 bits");
+
+    switch (opsize) {
+    case BITS64:
+        brcast_num = BITS64 / brsize;
+        break;
+    default:
+        brcast_num = (opsize / BITS128) * (BITS64 / brsize) * 2;
+        break;
+    }
+
+    return brcast_num;
+}
+
 static enum match_result matches(const struct itemplate *itemp,
                                  insn *instruction, int bits)
 {
@@ -2248,8 +2280,7 @@ static enum match_result matches(const struct itemplate *itemp,
             if (deco_brsize) {
                 template_opsize = (deco_brsize == BR_BITS32 ? BITS32 : BITS64);
                 /* calculate the proper number : {1to<brcast_num>} */
-                brcast_num = (itemp->opd[i] & SIZE_MASK) / BITS128 *
-                                BITS64 / template_opsize * 2;
+                brcast_num = get_broadcast_num(itemp->opd[i], template_opsize);
             } else {
                 template_opsize = 0;
             }
@@ -2272,12 +2303,12 @@ static enum match_result matches(const struct itemplate *itemp,
                 }
             } else if (is_broadcast &&
                        (brcast_num !=
-                        (8U << ((deco & BRNUM_MASK) >> BRNUM_SHIFT)))) {
+                        (2U << ((deco & BRNUM_MASK) >> BRNUM_SHIFT)))) {
                 /*
                  * broadcasting opsize matches but the number of repeated memory
                  * element does not match.
-                 * if 64b double precision float is broadcasted to zmm (512b),
-                 * broadcasting decorator must be {1to8}.
+                 * if 64b double precision float is broadcasted to ymm (256b),
+                 * broadcasting decorator must be {1to4}.
                  */
                 return MERR_BRNUMMISMATCH;
             }
