@@ -128,12 +128,12 @@ static const struct forwrefinfo *forwref;
 
 static struct preproc_ops *preproc;
 
-enum op_type {
-    op_normal,                  /* Preprocess and assemble */
-    op_preprocess,              /* Preprocess only */
-    op_depend,                  /* Generate dependencies */
-};
-static enum op_type operating_mode;
+#define OP_NORMAL           (1u << 0)
+#define OP_PREPROCESS       (1u << 1)
+#define OP_DEPEND           (1u << 2)
+
+static unsigned int operating_mode;
+
 /* Dependency flags */
 static bool depend_emit_phony = false;
 static bool depend_missing_ok = false;
@@ -199,7 +199,7 @@ static void nasm_fputs(const char *line, FILE * outfile)
 }
 
 /* Convert a struct tm to a POSIX-style time constant */
-static int64_t posix_mktime(struct tm *tm)
+static int64_t make_posix_time(struct tm *tm)
 {
     int64_t t;
     int64_t y = tm->tm_year;
@@ -253,9 +253,9 @@ static void define_macros_early(void)
     }
 
     if (gm_p)
-        posix_time = posix_mktime(&gm);
+        posix_time = make_posix_time(&gm);
     else if (lt_p)
-        posix_time = posix_mktime(&lt);
+        posix_time = make_posix_time(&lt);
     else
         posix_time = 0;
 
@@ -337,12 +337,11 @@ int main(int argc, char **argv)
 
     tolower_init();
 
-    nasm_init_malloc_error();
     offsets = raa_init();
     forwrefs = saa_init((int32_t)sizeof(struct forwrefinfo));
 
     preproc = &nasmpp;
-    operating_mode = op_normal;
+    operating_mode = OP_NORMAL;
 
     seg_init();
 
@@ -370,13 +369,11 @@ int main(int argc, char **argv)
     /* define some macros dependent of command-line */
     define_macros_late();
 
-    depend_ptr = (depend_file || (operating_mode == op_depend)) ? &depend_list : NULL;
+    depend_ptr = (depend_file || (operating_mode & OP_DEPEND)) ? &depend_list : NULL;
     if (!depend_target)
         depend_target = quote_for_make(outname);
 
-    switch (operating_mode) {
-    case op_depend:
-        {
+    if (operating_mode & OP_DEPEND) {
             char *line;
 
             if (depend_missing_ok)
@@ -389,11 +386,7 @@ int main(int argc, char **argv)
             while ((line = preproc->getline()))
                 nasm_free(line);
             preproc->cleanup(0);
-        }
-        break;
-
-    case op_preprocess:
-        {
+    } else if (operating_mode & OP_PREPROCESS) {
             char *line;
             char *file_name = NULL;
             int32_t prior_linnum = 0;
@@ -441,57 +434,51 @@ int main(int argc, char **argv)
             if (ofile && terminate_after_phase)
                 remove(outname);
             ofile = NULL;
+    }
+
+    if (operating_mode & OP_NORMAL) {
+        /*
+         * We must call ofmt->filename _anyway_, even if the user
+         * has specified their own output file, because some
+         * formats (eg OBJ and COFF) use ofmt->filename to find out
+         * the name of the input file and then put that inside the
+         * file.
+         */
+        ofmt->filename(inname, outname);
+
+        ofile = fopen(outname, (ofmt->flags & OFMT_TEXT) ? "w" : "wb");
+        if (!ofile)
+            nasm_error(ERR_FATAL | ERR_NOFILE,
+                       "unable to open output file `%s'", outname);
+
+        /*
+         * We must call init_labels() before ofmt->init() since
+         * some object formats will want to define labels in their
+         * init routines. (eg OS/2 defines the FLAT group)
+         */
+        init_labels();
+
+        ofmt->init();
+        dfmt = ofmt->current_dfmt;
+        dfmt->init();
+
+        assemble_file(inname, depend_ptr);
+
+        if (!terminate_after_phase) {
+            ofmt->cleanup(using_debug_info);
+            cleanup_labels();
+            fflush(ofile);
+            if (ferror(ofile))
+                nasm_error(ERR_NONFATAL|ERR_NOFILE,
+                           "write error on output file `%s'", outname);
         }
-        break;
 
-    case op_normal:
-        {
-            /*
-             * We must call ofmt->filename _anyway_, even if the user
-             * has specified their own output file, because some
-             * formats (eg OBJ and COFF) use ofmt->filename to find out
-             * the name of the input file and then put that inside the
-             * file.
-             */
-            ofmt->filename(inname, outname);
-
-            ofile = fopen(outname, (ofmt->flags & OFMT_TEXT) ? "w" : "wb");
-            if (!ofile) {
-                nasm_error(ERR_FATAL | ERR_NOFILE,
-                             "unable to open output file `%s'", outname);
-            }
-
-            /*
-             * We must call init_labels() before ofmt->init() since
-             * some object formats will want to define labels in their
-             * init routines. (eg OS/2 defines the FLAT group)
-             */
-            init_labels();
-
-            ofmt->init();
-            dfmt = ofmt->current_dfmt;
-            dfmt->init();
-
-            assemble_file(inname, depend_ptr);
-
-            if (!terminate_after_phase) {
-                ofmt->cleanup(using_debug_info);
-                cleanup_labels();
-                fflush(ofile);
-                if (ferror(ofile)) {
-                    nasm_error(ERR_NONFATAL|ERR_NOFILE,
-                               "write error on output file `%s'", outname);
-                }
-            }
-
-            if (ofile) {
-                fclose(ofile);
-                if (terminate_after_phase)
-                    remove(outname);
-                ofile = NULL;
-            }
+        if (ofile) {
+            fclose(ofile);
+            if (terminate_after_phase)
+                remove(outname);
+            ofile = NULL;
         }
-        break;
     }
 
     if (depend_list && !terminate_after_phase)
@@ -854,7 +841,7 @@ static bool process_arg(char *p, char *q)
 
         case 'e':       /* preprocess only */
         case 'E':
-            operating_mode = op_preprocess;
+            operating_mode = OP_PREPROCESS;
             break;
 
         case 'a':       /* assemble only - don't preprocess */
@@ -901,16 +888,21 @@ set_warning:
         case 'M':
             switch (p[2]) {
             case 0:
-                operating_mode = op_depend;
+                operating_mode = OP_DEPEND;
                 break;
             case 'G':
-                operating_mode = op_depend;
+                operating_mode = OP_DEPEND;
                 depend_missing_ok = true;
                 break;
             case 'P':
                 depend_emit_phony = true;
                 break;
             case 'D':
+                operating_mode = OP_NORMAL;
+                depend_file = q;
+                advance = true;
+                break;
+            case 'F':
                 depend_file = q;
                 advance = true;
                 break;
