@@ -1,6 +1,6 @@
 /* ----------------------------------------------------------------------- *
  *   
- *   Copyright 1996-2009 The NASM Authors - All Rights Reserved
+ *   Copyright 1996-2013 The NASM Authors - All Rights Reserved
  *   See the file AUTHORS included with the NASM distribution for
  *   the specific copyright holders.
  *
@@ -762,6 +762,9 @@ static void bin_out(int32_t segto, const void *data,
 
     switch (type) {
     case OUT_ADDRESS:
+    {
+        int asize = abs(size);
+
         if (segment != NO_SEG && !find_section_by_index(segment)) {
             if (segment % 2)
                 nasm_error(ERR_NONFATAL, "binary output format does not support"
@@ -773,12 +776,19 @@ static void bin_out(int32_t segto, const void *data,
         }
         if (s->flags & TYPE_PROGBITS) {
             if (segment != NO_SEG)
-                add_reloc(s, size, segment, -1L);
+                add_reloc(s, asize, segment, -1L);
             p = mydata;
-	    WRITEADDR(p, *(int64_t *)data, size);
-            saa_wbytes(s->contents, mydata, size);
+	    WRITEADDR(p, *(int64_t *)data, asize);
+            saa_wbytes(s->contents, mydata, asize);
         }
+
+        /*
+         * Reassign size with sign dropped, we will need it
+         * for section length calculation.
+         */
+        size = asize;
 	break;
+    }
 
     case OUT_RAWDATA:
         if (s->flags & TYPE_PROGBITS)
@@ -1464,8 +1474,8 @@ static void do_output_bin(void)
 }
 
 /* Generate Intel hex file output */
-static int write_ith_record(unsigned int len, uint16_t addr,
-			    uint8_t type, void *data)
+static void write_ith_record(unsigned int len, uint16_t addr,
+                             uint8_t type, void *data)
 {
     char buf[1+2+4+2+255*2+2+2];
     char *p = buf;
@@ -1484,10 +1494,7 @@ static int write_ith_record(unsigned int len, uint16_t addr,
 	p += sprintf(p, "%02X", dptr[i]);
     p += sprintf(p, "%02X\n", csum);
 
-    if (fwrite(buf, 1, p-buf, ofile) != (size_t)(p-buf))
-	return -1;
-
-    return 0;
+    nasm_write(buf, p-buf, ofile);
 }
 
 static void do_output_ith(void)
@@ -1538,8 +1545,8 @@ static void do_output_ith(void)
 }
 
 /* Generate Motorola S-records */
-static int write_srecord(unsigned int len,  unsigned int alen,
-			 uint32_t addr, uint8_t type, void *data)
+static void write_srecord(unsigned int len,  unsigned int alen,
+                          uint32_t addr, uint8_t type, void *data)
 {
     char buf[2+2+8+255*2+2+2];
     char *p = buf;
@@ -1572,10 +1579,7 @@ static int write_srecord(unsigned int len,  unsigned int alen,
 	p += sprintf(p, "%02X", dptr[i]);
     p += sprintf(p, "%02X\n", csum);
 
-    if (fwrite(buf, 1, p-buf, ofile) != (size_t)(p-buf))
-	return -1;
-
-    return 0;
+    nasm_write(buf, p-buf, ofile);
 }
 
 static void do_output_srec(void)
