@@ -1,6 +1,6 @@
 /* ----------------------------------------------------------------------- *
  *
- *   Copyright 1996-2010 The NASM Authors - All Rights Reserved
+ *   Copyright 1996-2014 The NASM Authors - All Rights Reserved
  *   See the file AUTHORS included with the NASM distribution for
  *   the specific copyright holders.
  *
@@ -137,7 +137,8 @@ struct Section {
     int32_t index;
     struct Reloc *head, **tail;
     uint32_t flags;             /* section flags */
-    char name[9];
+    char *name;
+    int32_t namepos;            /* Offset of name into the strings table */
     int32_t pos, relpos;
 };
 
@@ -215,7 +216,7 @@ static uint32_t strslen;
 static void coff_gen_init(void);
 static void coff_sect_write(struct Section *, const uint8_t *, uint32_t);
 static void coff_write(void);
-static void coff_section_header(char *, int32_t, int32_t, int32_t, int32_t, int, int32_t);
+static void coff_section_header(char *, int32_t, int32_t, int32_t, int32_t, int32_t, int, int32_t);
 static void coff_write_relocs(struct Section *);
 static void coff_write_symbols(void);
 
@@ -272,6 +273,7 @@ static void coff_cleanup(int debuginfo)
             sects[i]->head = sects[i]->head->next;
             nasm_free(r);
         }
+        nasm_free(sects[i]->name);
         nasm_free(sects[i]);
     }
     nasm_free(sects);
@@ -284,6 +286,7 @@ static void coff_cleanup(int debuginfo)
 static int coff_make_section(char *name, uint32_t flags)
 {
     struct Section *s;
+    size_t namelen;
 
     s = nasm_zalloc(sizeof(*s));
 
@@ -294,8 +297,20 @@ static int coff_make_section(char *name, uint32_t flags)
         s->index = def_seg;
     else
         s->index = seg_alloc();
-    strncpy(s->name, name, 8);
-    s->name[8] = '\0';
+    s->namepos = -1;
+    namelen = strlen(name);
+    if (namelen > 8) {
+        if (win32 || win64) {
+            s->namepos = strslen + 4;
+            saa_wbytes(strs, name, namelen + 1);
+            strslen += namelen + 1;
+        } else {
+            namelen = 8;
+        }
+    }
+    s->name = nasm_malloc(namelen + 1);
+    strncpy(s->name, name, namelen);
+    s->name[namelen] = '\0';
     s->flags = flags;
 
     if (nsects >= sectlen) {
@@ -337,9 +352,11 @@ static int32_t coff_section_names(char *name, int pass, int *bits)
     if (*p)
         *p++ = '\0';
     if (strlen(name) > 8) {
-        nasm_error(ERR_WARNING, "COFF section names limited to 8 characters:"
-              " truncating");
-        name[8] = '\0';
+        if (!win32 && !win64) {
+            nasm_error(ERR_WARNING,
+                       "COFF section names limited to 8 characters:  truncating");
+            name[8] = '\0';
+        }
     }
     flags = 0;
 
@@ -425,9 +442,42 @@ static int32_t coff_section_names(char *name, int pass, int *bits)
         sects[i]->flags &= align_and;
         sects[i]->flags |= align_or;
     } else if (pass == 1) {
-        if (flags)
-            nasm_error(ERR_WARNING, "section attributes ignored on"
-                  " redeclaration of section `%s'", name);
+        /* Check if any flags are specified */
+        if (flags) {
+            unsigned int align_flags = flags & IMAGE_SCN_ALIGN_MASK;
+
+            /* Warn if non-alignment flags differ */
+            if ((flags ^ sects[i]->flags) & ~IMAGE_SCN_ALIGN_MASK) {
+                nasm_error(ERR_WARNING, "section attributes ignored on"
+                    " redeclaration of section `%s'", name);
+            }
+            /* Check if alignment might be needed */
+            if (align_flags > IMAGE_SCN_ALIGN_1BYTES) {
+                unsigned int sect_align_flags = sects[i]->flags & IMAGE_SCN_ALIGN_MASK;
+
+                /* Compute the actual alignment */
+                unsigned int align = 1u << ((align_flags - IMAGE_SCN_ALIGN_1BYTES) >> 20);
+
+                /* Update section header as needed */
+                if (align_flags > sect_align_flags) {
+                    sects[i]->flags = (sects[i]->flags & ~IMAGE_SCN_ALIGN_MASK) | align_flags;
+                }
+                /* Check if not already aligned */
+                if (sects[i]->len % align) {
+                    unsigned int padding = (align - sects[i]->len) % align;
+                    /* We need to write at most 8095 bytes */
+                    char buffer[8095];
+                    if (sects[i]->flags & IMAGE_SCN_CNT_CODE) {
+                        /* Fill with INT 3 instructions */
+                        memset(buffer, 0xCC, padding);
+                    } else {
+                        memset(buffer, 0x00, padding);
+                    }
+                    saa_wbytes(sects[i]->data, buffer, padding);
+                    sects[i]->len += padding;
+                }
+            }
+        }
     }
 
     return sects[i]->index;
@@ -596,8 +646,9 @@ static void coff_out(int32_t segto, const void *data,
             nasm_error(ERR_PANIC, "OUT_RAWDATA with other than NO_SEG");
         coff_sect_write(s, data, size);
     } else if (type == OUT_ADDRESS) {
-        if (!(win64)) {
-            if (size != 4 && (segment != NO_SEG || wrt != NO_SEG)) {
+        int asize = abs(size);
+        if (!win64) {
+            if (asize != 4 && (segment != NO_SEG || wrt != NO_SEG)) {
                 nasm_error(ERR_NONFATAL, "COFF format does not support non-32-bit"
                       " relocations");
             } else {
@@ -614,25 +665,25 @@ static void coff_out(int32_t segto, const void *data,
                 }
                 p = mydata;
                 WRITELONG(p, *(int64_t *)data + fix);
-                coff_sect_write(s, mydata, size);
+                coff_sect_write(s, mydata, asize);
             }
         } else {
             int32_t fix = 0;
             p = mydata;
-            if (size == 8) {
+            if (asize == 8) {
                 if (wrt == imagebase_sect) {
                     nasm_error(ERR_NONFATAL, "operand size mismatch: 'wrt "
                                WRT_IMAGEBASE "' is a 32-bit operand");
                 }
                 fix = coff_add_reloc(s, segment, IMAGE_REL_AMD64_ADDR64);
                 WRITEDLONG(p, *(int64_t *)data + fix);
-                coff_sect_write(s, mydata, size);
+                coff_sect_write(s, mydata, asize);
             } else {
                 fix = coff_add_reloc(s, segment,
                         wrt == imagebase_sect ?	IMAGE_REL_AMD64_ADDR32NB:
                                                 IMAGE_REL_AMD64_ADDR32);
                 WRITELONG(p, *(int64_t *)data + fix);
-                coff_sect_write(s, mydata, size);
+                coff_sect_write(s, mydata, asize);
             }
         }
     } else if (type == OUT_REL2ADR) {
@@ -686,7 +737,7 @@ typedef struct tagString {
 
 static STRING *Exports = NULL;
 static struct Section *directive_sec;
-void AddExport(char *name)
+static void AddExport(char *name)
 {
     STRING *rvp = Exports, *newS;
 
@@ -914,7 +965,7 @@ static void coff_write(void)
      */
     vsize = 0L;
     for (i = 0; i < nsects; i++) {
-        coff_section_header(sects[i]->name, vsize, sects[i]->len,
+        coff_section_header(sects[i]->name, sects[i]->namepos, vsize, sects[i]->len,
                             sects[i]->pos, sects[i]->relpos,
                             sects[i]->nrelocs, sects[i]->flags);
         vsize += sects[i]->len;
@@ -937,7 +988,7 @@ static void coff_write(void)
     saa_fpwrite(strs, ofile);
 }
 
-static void coff_section_header(char *name, int32_t vsize,
+static void coff_section_header(char *name, int32_t namepos, int32_t vsize,
                                 int32_t datalen, int32_t datapos,
                                 int32_t relpos, int nrelocs, int32_t flags)
 {
@@ -945,8 +996,32 @@ static void coff_section_header(char *name, int32_t vsize,
 
     (void)vsize;
 
-    strncpy(padname, name, 8);
-    fwrite(padname, 8, 1, ofile);
+    if (namepos == -1) {
+        strncpy(padname, name, 8);
+        nasm_write(padname, 8, ofile);
+    } else {
+        /*
+         * If name is longer than 8 bytes, write '/' followed
+         * by offset into the strings table represented as
+         * decimal number.
+         */
+        namepos = namepos % 100000000;
+        padname[0] = '/';
+        padname[1] = '0' + (namepos / 1000000);
+        namepos = namepos % 1000000;
+        padname[2] = '0' + (namepos / 100000);
+        namepos = namepos % 100000;
+        padname[3] = '0' + (namepos / 10000);
+        namepos = namepos % 10000;
+        padname[4] = '0' + (namepos / 1000);
+        namepos = namepos % 1000;
+        padname[5] = '0' + (namepos / 100);
+        namepos = namepos % 100;
+        padname[6] = '0' + (namepos / 10);
+        namepos = namepos % 10;
+        padname[7] = '0' + (namepos);
+        nasm_write(padname, 8, ofile);
+    }
 
     fwriteint32_t(0,            ofile); /* Virtual size field - set to 0 or vsize */
     fwriteint32_t(0L,           ofile); /* RVA/offset - we ignore */
@@ -998,7 +1073,7 @@ static void coff_symbol(char *name, int32_t strpos, int32_t value,
 
     if (name) {
         strncpy(padname, name, 8);
-        fwrite(padname, 8, 1, ofile);
+        nasm_write(padname, 8, ofile);
     } else {
         fwriteint32_t(0, ofile);
         fwriteint32_t(strpos, ofile);
@@ -1022,7 +1097,7 @@ static void coff_write_symbols(void)
      */
     coff_symbol(".file", 0L, 0L, -2, 0, 0x67, 1);
     strncpy(filename, coff_infile, 18);
-    fwrite(filename, 18, 1, ofile);
+    nasm_write(filename, 18, ofile);
 
     /*
      * The section records, with their auxiliaries.
@@ -1033,7 +1108,7 @@ static void coff_write_symbols(void)
         coff_symbol(sects[i]->name, 0L, 0L, i + 1, 0, 3, 1);
         fwriteint32_t(sects[i]->len,    ofile);
         fwriteint16_t(sects[i]->nrelocs,ofile);
-        fwrite(filename, 12, 1, ofile);
+        nasm_write(filename, 12, ofile);
     }
 
     /*
