@@ -1,6 +1,6 @@
 /* ----------------------------------------------------------------------- *
  *
- *   Copyright 1996-2010 The NASM Authors - All Rights Reserved
+ *   Copyright 1996-2013 The NASM Authors - All Rights Reserved
  *   See the file AUTHORS included with the NASM distribution for
  *   the specific copyright holders.
  *
@@ -60,46 +60,8 @@
 
 #ifdef OF_ELF64
 
-/*
- * Relocation types.
- */
-struct Reloc {
-    struct Reloc *next;
-    int64_t address;            /* relative to _start_ of section */
-    int64_t symbol;             /* symbol index */
-    int64_t offset;             /* symbol addend */
-    int type;                   /* type of relocation */
-};
-
-struct Symbol {
-    struct rbtree symv;         /* symbol value and rbtree of globals */
-    int32_t strpos;             /* string table position of name */
-    int32_t section;            /* section ID of the symbol */
-    int type;                   /* symbol type */
-    int other;                  /* symbol visibility */
-    int32_t size;               /* size of symbol */
-    int32_t globnum;            /* symbol table offset if global */
-    struct Symbol *nextfwd;     /* list of unresolved-size symbols */
-    char *name;                 /* used temporarily if in above list */
-};
-
-struct Section {
-    struct SAA *data;
-    uint64_t len, size;
-    uint32_t nrelocs;
-    int32_t index;              /* index into sects array */
-    int type;                   /* SHT_PROGBITS or SHT_NOBITS */
-    uint64_t align;             /* alignment: power of two */
-    uint64_t flags;             /* section flags */
-    char *name;
-    struct SAA *rel;
-    uint64_t rellen;
-    struct Reloc *head, **tail;
-    struct rbtree *gsyms;       /* global symbols in section */
-};
-
 #define SECT_DELTA 32
-static struct Section **sects;
+static struct elf_section **sects;
 static int nsects, sectlen;
 
 #define SHSTR_DELTA 256
@@ -116,13 +78,10 @@ static struct RAA *bsym;
 static struct SAA *strs;
 static uint32_t strslen;
 
-static struct Symbol *fwds;
+static struct elf_symbol *fwds;
 
 static char elf_module[FILENAME_MAX];
 
-static uint8_t elf_osabi = 0;   /* Default OSABI = 0 (System V or Linux) */
-static uint8_t elf_abiver = 0;  /* Current ABI version */
-
 extern struct ofmt of_elf64;
 
 static struct ELF_SECTDATA {
@@ -134,13 +93,13 @@ static int elf_nsect, nsections;
 static int64_t elf_foffs;
 
 static void elf_write(void);
-static void elf_sect_write(struct Section *, const void *, size_t);
-static void elf_sect_writeaddr(struct Section *, int64_t, size_t);
+static void elf_sect_write(struct elf_section *, const void *, size_t);
+static void elf_sect_writeaddr(struct elf_section *, int64_t, size_t);
 static void elf_section_header(int, int, uint64_t, void *, bool, uint64_t, int, int,
                                int, int);
 static void elf_write_sections(void);
 static struct SAA *elf_build_symtab(int32_t *, int32_t *);
-static struct SAA *elf_build_reltab(uint64_t *, struct Reloc *);
+static struct SAA *elf_build_reltab(uint64_t *, struct elf_reloc *);
 static void add_sectname(char *, char *);
 
 struct erel {
@@ -198,7 +157,7 @@ static int64_t dwarf_infosym, dwarf_abbrevsym, dwarf_linesym;
 
 static struct dfmt df_dwarf;
 static struct dfmt df_stabs;
-static struct Symbol *lastsym;
+static struct elf_symbol *lastsym;
 
 /* common debugging routines */
 static void debug64_typevalue(int32_t);
@@ -235,7 +194,7 @@ static void elf_init(void)
     maxbits = 64;
     sects = NULL;
     nsects = sectlen = 0;
-    syms = saa_init((int32_t)sizeof(struct Symbol));
+    syms = saa_init((int32_t)sizeof(struct elf_symbol));
     nlocals = nglobs = ndebugs = 0;
     bsym = raa_init();
     strs = saa_init(1L);
@@ -267,7 +226,7 @@ static void elf_init(void)
 
 static void elf_cleanup(int debuginfo)
 {
-    struct Reloc *r;
+    struct elf_reloc *r;
     int i;
 
     (void)debuginfo;
@@ -306,7 +265,7 @@ static void add_sectname(char *firsthalf, char *secondhalf)
 
 static int elf_make_section(char *name, int type, int flags, int align)
 {
-    struct Section *s;
+    struct elf_section *s;
 
     s = nasm_zalloc(sizeof(*s));
 
@@ -351,8 +310,8 @@ static int32_t elf_section_names(char *name, int pass, int *bits)
         *p++ = '\0';
     flags_and = flags_or = type = align = 0;
 
-    section_attrib(name, p, pass, &flags_and,
-                   &flags_or, &align, &type);
+    elf_section_attrib(name, p, pass, &flags_and,
+                       &flags_or, &align, &type);
 
     if (!strcmp(name, ".shstrtab") ||
         !strcmp(name, ".symtab") ||
@@ -394,7 +353,7 @@ static void elf_deflabel(char *name, int32_t segment, int64_t offset,
                          int is_global, char *special)
 {
     int pos = strslen;
-    struct Symbol *sym;
+    struct elf_symbol *sym;
     bool special_used = false;
 
 #if defined(DEBUG) && DEBUG>2
@@ -416,7 +375,7 @@ static void elf_deflabel(char *name, int32_t segment, int64_t offset,
     }
 
     if (is_global == 3) {
-        struct Symbol **s;
+        struct elf_symbol **s;
         /*
          * Fix up a forward-reference symbol size from the first
          * pass.
@@ -606,12 +565,12 @@ static void elf_deflabel(char *name, int32_t segment, int64_t offset,
         nasm_error(ERR_NONFATAL, "no special symbol features supported here");
 }
 
-static void elf_add_reloc(struct Section *sect, int32_t segment,
+static void elf_add_reloc(struct elf_section *sect, int32_t segment,
                           int64_t offset, int type)
 {
-    struct Reloc *r;
+    struct elf_reloc *r;
 
-    r = *sect->tail = nasm_zalloc(sizeof(struct Reloc));
+    r = *sect->tail = nasm_zalloc(sizeof(struct elf_reloc));
     sect->tail = &r->next;
 
     r->address = sect->len;
@@ -652,13 +611,13 @@ static void elf_add_reloc(struct Section *sect, int32_t segment,
  * Inefficiency: we search, currently, using a linked list which
  * isn't even necessarily sorted.
  */
-static void elf_add_gsym_reloc(struct Section *sect,
+static void elf_add_gsym_reloc(struct elf_section *sect,
                                int32_t segment, uint64_t offset, int64_t pcrel,
                                int type, bool exact)
 {
-    struct Reloc *r;
-    struct Section *s;
-    struct Symbol *sym;
+    struct elf_reloc *r;
+    struct elf_section *s;
+    struct elf_symbol *sym;
     struct rbtree *srb;
     int i;
 
@@ -690,9 +649,9 @@ static void elf_add_gsym_reloc(struct Section *sect,
               " for this reference");
         return;
     }
-    sym = container_of(srb, struct Symbol, symv);
+    sym = container_of(srb, struct elf_symbol, symv);
 
-    r = *sect->tail = nasm_malloc(sizeof(struct Reloc));
+    r = *sect->tail = nasm_malloc(sizeof(struct elf_reloc));
     sect->tail = &r->next;
     r->next = NULL;
 
@@ -708,7 +667,7 @@ static void elf_out(int32_t segto, const void *data,
                     enum out_type type, uint64_t size,
                     int32_t segment, int32_t wrt)
 {
-    struct Section *s;
+    struct elf_section *s;
     int64_t addr;
     int reltype, bytes;
     int i;
@@ -785,6 +744,10 @@ static void elf_out(int32_t segto, const void *data,
 	break;
 
     case OUT_ADDRESS:
+    {
+        int isize = (int)size;
+        int asize = abs(size);
+
         addr = *(int64_t *)data;
         if (segment == NO_SEG) {
             /* Do nothing */
@@ -793,17 +756,23 @@ static void elf_out(int32_t segto, const void *data,
                   " segment base references");
         } else {
             if (wrt == NO_SEG) {
-                switch ((int)size) {
+                switch (isize) {
                 case 1:
+                case -1:
                     elf_add_reloc(s, segment, addr, R_X86_64_8);
                     break;
                 case 2:
+                case -2:
                     elf_add_reloc(s, segment, addr, R_X86_64_16);
                     break;
                 case 4:
                     elf_add_reloc(s, segment, addr, R_X86_64_32);
                     break;
+                case -4:
+                    elf_add_reloc(s, segment, addr, R_X86_64_32S);
+                    break;
                 case 8:
+                case -8:
                     elf_add_reloc(s, segment, addr, R_X86_64_64);
                     break;
                 default:
@@ -821,7 +790,7 @@ static void elf_out(int32_t segto, const void *data,
                 elf_add_reloc(s, segment, addr, R_X86_64_GOTPC32);
                 addr = 0;
             } else if (wrt == elf_gotoff_sect + 1) {
-                if (size != 8) {
+                if (asize != 8) {
                     nasm_error(ERR_NONFATAL, "ELF64 requires ..gotoff "
                           "references to be qword");
                 } else {
@@ -829,7 +798,7 @@ static void elf_out(int32_t segto, const void *data,
                     addr = 0;
                 }
             } else if (wrt == elf_got_sect + 1) {
-                switch ((int)size) {
+                switch (asize) {
                 case 4:
                     elf_add_gsym_reloc(s, segment, addr, 0,
                                        R_X86_64_GOT32, true);
@@ -845,13 +814,15 @@ static void elf_out(int32_t segto, const void *data,
                     break;
                 }
             } else if (wrt == elf_sym_sect + 1) {
-                switch ((int)size) {
+                switch (isize) {
                 case 1:
+                case -1:
                     elf_add_gsym_reloc(s, segment, addr, 0,
                                        R_X86_64_8, false);
                     addr = 0;
                     break;
                 case 2:
+                case -2:
                     elf_add_gsym_reloc(s, segment, addr, 0,
                                        R_X86_64_16, false);
                     addr = 0;
@@ -861,7 +832,13 @@ static void elf_out(int32_t segto, const void *data,
                                        R_X86_64_32, false);
                     addr = 0;
                     break;
+                case -4:
+                    elf_add_gsym_reloc(s, segment, addr, 0,
+                                       R_X86_64_32S, false);
+                    addr = 0;
+                    break;
                 case 8:
+                case -8:
                     elf_add_gsym_reloc(s, segment, addr, 0,
                                        R_X86_64_64, false);
                     addr = 0;
@@ -878,8 +855,9 @@ static void elf_out(int32_t segto, const void *data,
                       " use of WRT");
             }
         }
-        elf_sect_writeaddr(s, addr, size);
+        elf_sect_writeaddr(s, addr, asize);
 	break;
+    }
 
     case OUT_REL1ADR:
 	reltype = R_X86_64_PC8;
@@ -1047,7 +1025,7 @@ static void elf_write(void)
     /*
      * Output the ELF header.
      */
-    fwrite("\177ELF\2\1\1", 7, 1, ofile);
+    nasm_write("\177ELF\2\1\1", 7, ofile);
     fputc(elf_osabi, ofile);
     fputc(elf_abiver, ofile);
     fwritezero(7, ofile);
@@ -1204,7 +1182,7 @@ static void elf_write(void)
 static struct SAA *elf_build_symtab(int32_t *len, int32_t *local)
 {
     struct SAA *s = saa_init(1L);
-    struct Symbol *sym;
+    struct elf_symbol *sym;
     uint8_t entry[24], *p;
     int i;
 
@@ -1323,7 +1301,7 @@ static struct SAA *elf_build_symtab(int32_t *len, int32_t *local)
     return s;
 }
 
-static struct SAA *elf_build_reltab(uint64_t *len, struct Reloc *r)
+static struct SAA *elf_build_reltab(uint64_t *len, struct elf_reloc *r)
 {
     struct SAA *s;
     uint8_t *p, entry[24];
@@ -1396,18 +1374,18 @@ static void elf_write_sections(void)
             if (elf_sects[i].is_saa)
                 saa_fpwrite(elf_sects[i].data, ofile);
             else
-                fwrite(elf_sects[i].data, len, 1, ofile);
+                nasm_write(elf_sects[i].data, len, ofile);
             fwritezero(align, ofile);
         }
 }
 
-static void elf_sect_write(struct Section *sect, const void *data, size_t len)
+static void elf_sect_write(struct elf_section *sect, const void *data, size_t len)
 {
     saa_wbytes(sect->data, data, len);
     sect->len += len;
 }
 
-static void elf_sect_writeaddr(struct Section *sect, int64_t data, size_t len)
+static void elf_sect_writeaddr(struct elf_section *sect, int64_t data, size_t len)
 {
     saa_writeaddr(sect->data, data, len);
     sect->len += len;
@@ -1415,7 +1393,7 @@ static void elf_sect_writeaddr(struct Section *sect, int64_t data, size_t len)
 
 static void elf_sectalign(int32_t seg, unsigned int value)
 {
-    struct Section *s = NULL;
+    struct elf_section *s = NULL;
     int i;
 
     for (i = 0; i < nsects; i++) {
@@ -1436,46 +1414,6 @@ static int32_t elf_segbase(int32_t segment)
     return segment;
 }
 
-static int elf_directive(enum directives directive, char *value, int pass)
-{
-    bool err;
-    int64_t n;
-    char *p;
-
-    switch (directive) {
-    case D_OSABI:
-        if (pass == 2)
-            return 1; /* ignore in pass 2 */
-
-        n = readnum(value, &err);
-        if (err) {
-            nasm_error(ERR_NONFATAL, "`osabi' directive requires a parameter");
-            return 1;
-        }
-        if (n < 0 || n > 255) {
-            nasm_error(ERR_NONFATAL, "valid osabi numbers are 0 to 255");
-            return 1;
-        }
-        elf_osabi  = n;
-        elf_abiver = 0;
-
-        if ((p = strchr(value,',')) == NULL)
-            return 1;
-
-        n = readnum(p+1, &err);
-        if (err || n < 0 || n > 255) {
-            nasm_error(ERR_NONFATAL, "invalid ABI version number (valid: 0 to 255)");
-            return 1;
-        }
-
-        elf_abiver = n;
-        return 1;
-
-    default:
-        return 0;
-    }
-}
-
 static void elf_filename(char *inname, char *outname)
 {
     strcpy(elf_module, inname);
