@@ -1,6 +1,6 @@
 /* ----------------------------------------------------------------------- *
  *   
- *   Copyright 1996-2009 The NASM Authors - All Rights Reserved
+ *   Copyright 1996-2014 The NASM Authors - All Rights Reserved
  *   See the file AUTHORS included with the NASM distribution for
  *   the specific copyright holders.
  *
@@ -1086,6 +1086,9 @@ static void obj_out(int32_t segto, const void *data,
     {
         int rsize;
 
+        if (type == OUT_ADDRESS)
+            size = abs(size);
+
         if (segment == NO_SEG && type != OUT_ADDRESS)
             nasm_error(ERR_NONFATAL, "relative call to absolute address not"
                   " supported by OBJ format");
@@ -2393,7 +2396,7 @@ static void obj_fwrite(ObjRecord * orp)
     len = orp->committed + 1;
     cksum += (len & 0xFF) + ((len >> 8) & 0xFF);
     fwriteint16_t(len, ofile);
-    fwrite(orp->buf, 1, len - 1, ofile);
+    nasm_write(orp->buf, len-1, ofile);
     for (ptr = orp->buf; --len; ptr++)
         cksum += *ptr;
     fputc((-cksum) & 0xFF, ofile);
@@ -2401,7 +2404,7 @@ static void obj_fwrite(ObjRecord * orp)
 
 extern macros_t obj_stdmac[];
 
-void dbgbi_init(void)
+static void dbgbi_init(void)
 {
     fnhead = NULL;
     fntail = &fnhead;
